// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <memory>
#include <cstring>
#include "InternalDmaFilterSession.h"
#include "InternalFrame.h"
#include "InternalSideData.h"
#include "vpi_memory.h"
#include "vpi_device.h"

namespace xma {

#define XMA_DMA_FILTER_SESSION "DmaFilterSession"

///////////////////////////////////////////////////////////////////////////////
InternalDmaFilterSession::InternalDmaFilterSession(IntSessionFactoryPtr sf, XmaFilterProperties* props) :
    FilterThread<>(sf, "dma_filter", props->params, props->param_cnt, 0, 1) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession()");
  if (props->input.sw_format != props->output.sw_format) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Input sw format must be the same as output sw format!");
  }
  if ((props->input.flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) || (props->output.flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Tile flag not supported!");
  }
  if ((props->input.flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) || (props->output.flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Compress flag not supported!");
  }
  if (props->hwfilter_type == XMA_UPLOAD_FILTER_TYPE) {
    if ((props->input.sw_format != props->input.format) && (props->input.format != XMA_VPE_FMT_TYPE)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Invalid input format!");
    }
    if (props->output.format != XMA_VPE_FMT_TYPE) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Output format must be on device!");
    }
    m_dir                             = VPI_MEMORY_TRANS_RC2EP;
    IntFramePropertiesPtr frame_props = std::make_shared<InternalFrameProperties>(
        m_sf, m_log, XMA_VPE_FMT_TYPE, props->input.sw_format, props->input.width, props->input.height, 0);
    m_pool = std::make_shared<BufferPool>(m_sf->GetMemHandle(), m_log, XMA_DMA_DEFAULT_DELAY, InternalFrameProperties::GetBufferSize(frame_props), EP_SIDE_EN);
    if (!m_pool) {
      Close();
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Out of memory!");
    }
  } else if (props->hwfilter_type == XMA_DOWNLOAD_FILTER_TYPE) {
    if ((props->input.sw_format != props->input.format) && (props->input.format != XMA_VPE_FMT_TYPE)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Invalid input format!");
    }
    if (props->output.format == XMA_VPE_FMT_TYPE) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Output format must be on host!");
    }
    m_dir = VPI_MEMORY_TRANS_EP2RC;
  } else {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() Invalid filter type!");
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::InternalDmaFilterSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalDmaFilterSession::~InternalDmaFilterSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::~InternalDmaFilterSession()");

  Close();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDmaFilterSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Enqueue()");

  int32_t ret = FilterThread::Push(data);

  switch (ret) {
  case XMA_TRY_AGAIN:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Enqueue() - Returning try again");
    break;
  case XMA_SUCCESS:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Enqueue() - Returning success");
    break;
  default:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Enqueue() - Returning");
    break;
  }
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDmaFilterSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Dequeue()");

  int32_t ret = FilterThread::Pull(data);

  char error_msg[1024];
  sprintf(error_msg, "InternalDmaFilterSession::Dequeue() - returning %d", ret);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, error_msg);

  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDmaFilterSession::Process(queue_t& in, queue_t& out, EmptyDynamicParams*) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Process()");
  FramePtr frame = std::get<QueueFrame>(in).frame;
  if (!frame) {
    out = QueueFrame();
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Process() - Returning EOS");
    return XMA_EOS;
  }

  FramePropertiesPtr props = frame->GetProps();
  if (((props->GetFormat() == XMA_VPE_FMT_TYPE) && (frame->GetBufferType() == XMA_HOST_BUFFER_TYPE)) ||
      ((props->GetFormat() != XMA_VPE_FMT_TYPE) && (frame->GetBufferType() != XMA_HOST_BUFFER_TYPE)) || (frame->GetBufferType() == NO_BUFFER)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Process() Inconsistent buffer types!");
  }

  FramePtr new_frame;
  if (m_dir == VPI_MEMORY_TRANS_RC2EP) {
    new_frame = CopyToDevice(frame);
  } else {
    new_frame = CopyToHost(frame);
  }

  out                             = QueueFrame();
  std::get<QueueFrame>(out).frame = new_frame;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::Process() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
FramePtr InternalDmaFilterSession::CopyToHost(FramePtr src) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost()");
  FramePropertiesPtr props = src->GetProps();

  if (props->GetFormat() != XMA_VPE_FMT_TYPE) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() - Returning already on host");
    return src;
  }

  if ((props->GetFlags() & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) != 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() - Unable to DMA from device to host, frame is tiled!");
  }

  if ((props->GetFlags() & XMA_FRAME_PROPERTY_FLAG_COMPRESS) != 0) {
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() - Unable to DMA from device to host, frame is compressed!");
  }

  IntFramePropertiesPtr new_props = std::make_shared<InternalFrameProperties>(
      m_sf, m_log, props->GetSwFormat(), props->GetSwFormat(), props->GetWidth(), props->GetHeight(), props->GetFlags());
  buffer_t    buffer    = m_sf->GetBuffer(InternalFrameProperties::GetBufferSize(new_props));
  IntFramePtr new_frame = std::make_shared<InternalFrame>(m_sf, m_log, new_props, false, std::move(buffer), src->GetFrameRate(), src->GetPts(), src->IsIdr());

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    VpiLinearMem* mem = static_cast<VpiLinearMem*>(src->GetBuffer(plane));
    if ((!mem) || (!mem->bus_address_ep)) {
      THROW_RUNTIME_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() Unable to DMA from device to host, no device buffer!");
    }
    VpiLinearMem temp_mem;
    memset(&temp_mem, 0, sizeof(VpiLinearMem));
    temp_mem.size               = new_props->GetPlaneSize(plane);
    temp_mem.request_size       = temp_mem.size;
    temp_mem.virtual_address_rc = new_frame->GetHostBuffer(plane);
    if (vpi_memory_mem_sync(m_sf->GetMemHandle(), mem, &temp_mem, 0, props->GetPlaneSize(plane), m_dir) < 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() DMA failed!");
    }
  }

  IntFramePtr int_src = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_RUNTIME_EXCEPTION(m_log,
        XMA_ERROR_LOG,
        XMA_DMA_FILTER_SESSION,
        "InternalDmaFilterSession::CopyToHost() Frame not allocated with SessionFactory::AllocFrame() or SessionFactory::CloneFrame()!");
  }
  SideDataPtr src_sd;
  for (uint32_t index = 0; (src_sd = int_src->GetSideDataAtIndex(index)); ++index) {
    XmaBufferType buffer_type = src_sd->GetBufferType();
    if ((buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE)) {
      new_frame->AddSideData(src_sd);
    } else {
      size_t         size       = src_sd->GetSize();
      IntSideDataPtr int_src_sd = std::dynamic_pointer_cast<InternalSideData>(src_sd);
      if (!int_src_sd) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() invalid source side data!");
      }
      side_data_buffer_t src_buffer = int_src_sd->GetVpiSideData();
      VpiLinearMem       src_mem;
      memset(&src_mem, 0, sizeof(VpiLinearMem));
      src_mem.size                  = size;
      src_mem.request_size          = size;
      src_mem.bus_address_ep        = reinterpret_cast<addr_t>(vpi_get_sidedata_paddr(src_buffer.get()));
      IntSideDataPtr     new_sd     = InternalSideData::AllocSideData(m_sf, m_log, src_sd->GetType(), XMA_HOST_BUFFER_TYPE, size);
      side_data_buffer_t new_buffer = new_sd->GetVpiSideData();
      VpiLinearMem       new_mem;
      memset(&new_mem, 0, sizeof(VpiLinearMem));
      new_mem.size               = size;
      new_mem.request_size       = size;
      new_mem.virtual_address_rc = vpi_get_sidedata_vaddr(new_buffer.get());
      if (vpi_memory_mem_sync(m_sf->GetMemHandle(), &src_mem, &new_mem, 0, size, m_dir) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() DMA failed!");
      }
      new_sd->CopyMetadata(src_sd);
      new_frame->AddSideData(new_sd);
      new_sd->DecRef(false);
    }
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToHost() - Returning");
  return new_frame;
}

///////////////////////////////////////////////////////////////////////////////
FramePtr InternalDmaFilterSession::CopyToDevice(FramePtr src) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice()");

  FramePropertiesPtr    props     = src->GetProps();
  IntFramePropertiesPtr new_props = std::make_shared<InternalFrameProperties>(
      m_sf, m_log, XMA_VPE_FMT_TYPE, props->GetSwFormat(), props->GetWidth(), props->GetHeight(), props->GetFlags());
  if ((m_temp_props) && ((m_temp_props->GetWidth() != props->GetWidth()) || (m_temp_props->GetHeight() != props->GetHeight()))) {
    uint32_t num = m_pool->GetNum();
    m_pool       = std::make_shared<BufferPool>(m_sf->GetMemHandle(), m_log, num, InternalFrameProperties::GetBufferSize(new_props), EP_SIDE_EN);
    if (!m_pool) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() Out of memory!");
    }
    m_temp_props.reset();
    m_temp.reset();
  }
  IntFramePtr new_frame = std::make_shared<InternalFrame>(
      m_sf, m_log, new_props, false, std::move(m_pool->GetBuffer()), src->GetFrameRate(), src->GetPts(), src->IsIdr());

  int32_t planes = props->GetNumPlanes();
  int32_t plane;
  for (plane = 0; plane < planes; ++plane) {
    VpiLinearMem* mem = nullptr;
    VpiLinearMem  mem0;
    if (props->GetFormat() == XMA_VPE_FMT_TYPE) {
      mem = static_cast<VpiLinearMem*>(src->GetBuffer(plane));
      if ((!mem) || (!mem->virtual_address_rc)) {
        THROW_RUNTIME_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() Unable to DMA from host to device, no host buffer!");
      }
    } else {
      mem = &mem0;
      memset(mem, 0, sizeof(VpiLinearMem));
      mem->size               = props->GetPlaneSize(plane);
      mem->request_size       = mem->size;
      mem->virtual_address_rc = src->GetHostBuffer(plane);
    }

    if (props->GetPlaneSize(plane) == new_props->GetPlaneSize(plane)) {
      if (vpi_memory_mem_sync(m_sf->GetMemHandle(), mem, static_cast<VpiLinearMem*>(new_frame->GetBuffer(plane)), 0, props->GetPlaneSize(plane), m_dir) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() DMA failed!");
      }
    } else if (props->GetPlaneStride(plane) == new_props->GetPlaneStride(plane)) {
      if (vpi_memory_mem_sync(
              m_sf->GetMemHandle(), mem, static_cast<VpiLinearMem*>(new_frame->GetBuffer(plane)), 0, props->GetPlaneStride(plane) * props->GetHeight(), m_dir) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() DMA failed!");
      }
    } else {
      int32_t width  = props->GetWidth();
      int32_t height = props->GetHeight();

      if ((!m_temp_props.get()) || (!m_temp.get())) {
        m_temp_props    = std::make_shared<InternalFrameProperties>(m_sf, m_log, props->GetSwFormat(), props->GetSwFormat(), width, height, props->GetFlags());
        buffer_t buffer = m_sf->GetBuffer(InternalFrameProperties::GetBufferSize(m_temp_props));
        m_temp          = std::make_shared<InternalFrame>(m_sf, m_log, m_temp_props, false, std::move(buffer));
      }
      size_t   src_stride = props->GetPlaneStride(plane);
      size_t   dst_stride = m_temp_props->GetPlaneStride(plane);
      size_t   min_stride = (src_stride <= dst_stride) ? src_stride : dst_stride;
      int32_t  src_height = props->GetPlaneHeight(plane);
      int32_t  dst_height = m_temp_props->GetPlaneHeight(plane);
      int32_t  min_height = (src_height <= dst_height) ? src_height : dst_height;
      uint8_t* psrc       = src->GetHostBuffer(plane);
      uint8_t* pdst       = m_temp->GetHostBuffer(plane);
      int32_t  y;
      for (y = 0; y < min_height; ++y) {
        memcpy(pdst, psrc, min_stride);
        psrc += src_stride;
        pdst += dst_stride;
      }
      mem->size               = m_temp_props->GetPlaneSize(plane);
      mem->request_size       = mem->size;
      mem->virtual_address_rc = m_temp->GetHostBuffer(plane);
      if (vpi_memory_mem_sync(m_sf->GetMemHandle(), mem, static_cast<VpiLinearMem*>(new_frame->GetBuffer(plane)), 0, props->GetPlaneSize(plane), m_dir) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() DMA failed!");
      }
    }
  }

  IntFramePtr int_src = std::dynamic_pointer_cast<InternalFrame>(src);
  if (!int_src) {
    THROW_RUNTIME_EXCEPTION(m_log,
        XMA_ERROR_LOG,
        XMA_DMA_FILTER_SESSION,
        "InternalDmaFilterSession::CopyToHost() Frame not allocated with SessionFactory::AllocFrame() or SessionFactory::CloneFrame()!");
  }
  SideDataPtr src_sd;
  for (uint32_t index = 0; (src_sd = int_src->GetSideDataAtIndex(index)); ++index) {
    XmaBufferType buffer_type = src_sd->GetBufferType();
    if ((buffer_type == XMA_DEVICE_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE)) {
      new_frame->AddSideData(src_sd);
    } else {
      size_t         size   = src_sd->GetSize();
      IntSideDataPtr new_sd = InternalSideData::AllocSideData(m_sf, m_log, src_sd->GetType(), XMA_DEVICE_BUFFER_TYPE, size);
      memcpy(new_sd->GetHostPtr(), src_sd->GetHostPtr(), size);
      new_sd->Write();
      new_sd->CopyMetadata(src_sd);
      new_frame->AddSideData(new_sd);
      new_sd->DecRef(false);
    }
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DMA_FILTER_SESSION, "InternalDmaFilterSession::CopyToDevice() - Returning");
  return new_frame;
}

} // namespace xma
