// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <memory>
#include <cstring>
#include <string.h>
#include <unistd.h>
#include "InternalEncoderSession.h"
#include "InternalFrame.h"
#include "InternalDataBuffer.h"
#include "xma.h"
#include "vpi_pic.h"

#include "log_ama.h"

#ifndef MAX
#define MAX(a, b) (((a) > (b)) ? (a) : (b))
#endif

typedef struct VpiEncContext {
  VpiEncType vpi_encoder_type;
  void*      vpi_encoder_handle;
} VpiEncContext;

// Do *NOT* modify.  This must match an internal structure layout
typedef struct _Xav1Context {
  VpiEncInitParams props;
  VpiEncHandle     xav1enc;
  int              poc;
  int              max_input_delay;
  int              max_frames_delay;
  bool             isInputPoolExtended;
  void*            extradata;
  uint32_t         log_frm_num;
} Xav1Context;

struct EnumName {
  int32_t     enum_value;
  const char* enum_name;
};

EnumName PresetList[] = {
    {XMA_ENC_PRESET_MEDIUM, "medium"},
    {XMA_ENC_PRESET_SLOW, "slow"},
    {XMA_ENC_PRESET_FAST, "fast"},
};

EnumName H264ProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_H264_BASELINE, "baseline"},
    {XMA_ENC_PROFILE_H264_MAIN, "main"},
    {XMA_ENC_PROFILE_H264_HIGH, "high"},
    {XMA_ENC_PROFILE_H264_HIGH10, "high-10"},
    {XMA_ENC_PROFILE_H264_HIGH10_INTRA, "high-10-intra"},
};

EnumName HEVCProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_HEVC_MAIN, "main"},
    {XMA_ENC_PROFILE_HEVC_MAIN_INTRA, "main-intra"},
    {XMA_ENC_PROFILE_HEVC_MAIN10, "main-10"},
    {XMA_ENC_PROFILE_HEVC_MAIN10_INTRA, "main-10-intra"},
};

EnumName AV1ProfileList[] = {
    {XMA_ENC_PROFILE_AUTO, "auto"},
    {XMA_ENC_PROFILE_AV1_MAIN, "main"},
};

const char* GetEnumName(EnumName arr[], int32_t count, int32_t value) {
  for (int32_t index = 0; index < count; ++index) {
    if (arr[index].enum_value == value) {
      return arr[index].enum_name;
    }
  }
  return NULL;
}

namespace xma {

#define XMA_ENCODER_SESSION "EncoderSession"

///////////////////////////////////////////////////////////////////////////////
InternalEncoderSession::InternalEncoderSession(IntSessionFactoryPtr sf, XmaEncoderProperties* props) : SessionBase(sf, "enc", props->params, props->param_cnt) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InternalEncoderSession()");

  snprintf(m_instance_name, 64, "encoder_@%p", this);
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);

  InitParams(props);

  if (SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_DELAY_INITIALIZATION, 0, 1, XMA_ENC_DELAY_INITIALIZATION_DEFAULT) == 0) {
    FramePtr frame;
    StartEncoder(frame);
  }

  if (m_use_thread) {
    m_thread = std::thread(InternalEncoderSession::Thread, this);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InternalEncoderSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalEncoderSession::~InternalEncoderSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::~InternalEncoderSession()");

  if (m_use_thread) {
    m_stop = true;
    m_thread_enqueue_cond.notify_all();
    m_thread_dequeue_cond.notify_all();
    m_wait_enqueue_cond.notify_all();
    m_wait_dequeue_cond.notify_all();
    m_thread.join();
  }
  m_handle.reset();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::SetLog(LogPtr log) {
  lock_t guard(m_mutex);
  SessionBase::SetLog(log);
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Enqueue(queue_t& data) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue()");
  FramePtr frame = std::move(std::get<QueueFrame>(data).frame);

  if (frame) {
    LogLatencySend(m_frame_enqueued + 1);
    IntFramePtr int_frame = std::dynamic_pointer_cast<InternalFrame>(frame);
    if (int_frame) {
      int_frame->SetFrozen();

      VpiBufferPool* vpi_bp = int_frame->GetBufferPool();
      uint32_t       depth  = vpi_buffer_pool_get_depth(vpi_bp);
      if (depth > m_input_bufferpool_size) {
        m_input_bufferpool_size = depth;
      } else if (depth < m_input_bufferpool_size) {
        int32_t extend_size = m_input_bufferpool_size - depth;
        if (extend_size > 0) {
          int_frame->ExtendBufferPool(extend_size);
        }
      }
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() frame not valid!");
    }
  }

  uint32_t api_version = m_sf->GetAPIVersion();
  int32_t  ret;
  if (api_version < XMA_API_VERSION_1_1) {
    ret = Push_v1_0(frame);
  } else {
    ret = Push_v1_1(frame);
  }
  if (ret == XMA_SUCCESS) {
    ++m_frame_enqueued;
  }
  char ret_str[64];
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Enqueue() - Returning %s", xma_perror(ret, ret_str, sizeof(ret_str)));
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Dequeue(queue_t& data) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue()");

  uint32_t         api_version = m_sf->GetAPIVersion();
  int32_t          ret;
  IntDataBufferPtr buf;
  if (api_version < XMA_API_VERSION_1_1) {
    ret = Pull_v1_0(buf);
  } else {
    ret = Pull_v1_1(buf);
  }
  if (ret == XMA_SUCCESS) {
    data                                 = QueueDataBuffer();
    std::get<QueueDataBuffer>(data).data = std::move(buf);
    --m_frames_processing;
    LogLatencyRecv(++m_frame_dequeued);
  }

  char ret_str[64];
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Dequeue() - Returning %s", xma_perror(ret, ret_str, sizeof(ret_str)));
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::InitParams(XmaEncoderProperties* props) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Failed to create VPI memory!");
  }

  XmaAPIVersion version = m_sf->GetAPIVersion();

  memset(&m_params, 0, sizeof(VpiEncInitParams));
  m_params.enc_type     = VPE_ENCODER_XAV1;
  m_params.hw_ctx       = m_sf->GetContext();
  m_params.mem_ctx      = m_mem_ctx.get();
  m_params.rate_num     = props->framerate.numerator;
  m_params.rate_denom   = props->framerate.denominator;
  m_params.width        = props->width;
  m_params.height       = props->height;
  m_params.walign       = XMA_FRAME_W_ALIGN;
  m_params.halign       = XMA_FRAME_H_ALIGN;
  m_params.bit_rate     = props->bitrate;
  m_params.qp           = props->qp;
  m_params.gop_size     = props->gop_size;
  m_params.rate_control = static_cast<RateControlMode>(props->rc_mode);
  m_params.enc_level    = props->level;
  m_params.enc_profile  = static_cast<EncProfile>(props->profile);

  if (props->lookahead_depth <= XMA_ENC_MAX_LOOKAHEAD_DEPTH) {
    m_params.la_depth = props->lookahead_depth;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for lookahead_depth!");
  }

  if (props->spat_aq_gain <= XMA_ENC_MAX_SPATIAL_AQ_GAIN) {
    m_params.spatial_aq_gain = props->spat_aq_gain;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for spatial_aq_gain!");
  }

  if (props->temp_aq_gain <= XMA_ENC_MAX_TEMPORAL_AQ_GAIN) {
    m_params.temporal_aq_gain = props->temp_aq_gain;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for temporal_aq_gain!");
  }

  if (version < XMA_API_VERSION_1_1) {
    if ((props->preset < XMA_ENC_PRESET_MEDIUM) || (props->preset > XMA_ENC_PRESET_SLOW)) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unsupported preset!");
    }
  } else {
    if ((props->preset < XMA_ENC_PRESET_FAST) || (props->preset > XMA_ENC_PRESET_SLOW)) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unsupported preset!");
    }
  }
  m_params.preset = const_cast<char*>(GetEnumName(PresetList, sizeof(PresetList) / sizeof(PresetList[0]), props->preset));

  int32_t enc_max_val_min_qp, enc_max_val_max_qp = 0;

  if (props->hwencoder_type == XMA_AV1_ENCODER_TYPE) {
    enc_max_val_min_qp = XMA_ENC_AV1_MIN_QP_MAX_VALUE;
    enc_max_val_max_qp = XMA_ENC_AV1_MAX_QP_MAX_VALUE;
  } else {
    enc_max_val_min_qp = XMA_ENC_MIN_QP_MAX_VALUE;
    enc_max_val_max_qp = XMA_ENC_MAX_QP_MAX_VALUE;
  }

  if (props->minQP >= XMA_ENC_QP_MIN_VALUE && props->minQP <= enc_max_val_min_qp) {
    m_params.min_qp = props->minQP;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for min_qp!");
  }

  if (props->maxQP >= XMA_ENC_QP_MIN_VALUE && props->maxQP <= enc_max_val_max_qp) {
    m_params.max_qp = props->maxQP;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for max_qp!");
  }

  XmaParameter* param = nullptr;
  VpiHwId       hw_id;
  switch (props->hwencoder_type) {
  case XMA_H264_ENCODER_TYPE:
    hw_id                 = VPE_HW_H264ENC;
    m_params.codec_format = VPE_ENC_H264;
    m_params.profile      = const_cast<char*>(GetEnumName(H264ProfileList, sizeof(H264ProfileList) / sizeof(H264ProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown H.264 profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 10:
    case 11:
    case 12:
    case 13:
    case 20:
    case 21:
    case 22:
    case 30:
    case 31:
    case 32:
    case 40:
    case 41:
    case 42:
    case 50:
    case 51:
    case 52:
    case 60:
    case 61:
    case 62:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown H.264 level!");
    }
    break;
  case XMA_HEVC_ENCODER_TYPE:
    hw_id                 = VPE_HW_HEVCENC;
    m_params.codec_format = VPE_ENC_HEVC;
    m_params.profile      = const_cast<char*>(GetEnumName(HEVCProfileList, sizeof(HEVCProfileList) / sizeof(HEVCProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown HEVC profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 10:
    case 20:
    case 21:
    case 30:
    case 31:
    case 40:
    case 41:
    case 50:
    case 51:
    case 52:
    case 60:
    case 61:
    case 62:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown HEVC level!");
    }
    break;
  case XMA_AV1_ENCODER_TYPE:
    if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_DEVICE_TYPE)) != nullptr) {
      switch (*static_cast<XmaEncoderDeviceType*>(param->value)) {
      case XMA_ENC_DEVICE_TYPE_ANY:
      case XMA_ENC_DEVICE_TYPE_1:
        hw_id                 = VPE_HW_XAV1ENC;
        m_params.codec_format = VPE_ENC_AV1;
        m_params.av1_type     = VPI_ENC_TYPE_1;
        break;
      case XMA_ENC_DEVICE_TYPE_2:
        hw_id                 = VPE_HW_XAV1ENC;
        m_params.codec_format = VPE_ENC_AV1_T2;
        m_params.av1_type     = VPI_ENC_TYPE_2;
        break;
      default:
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid device type!");
      }
    } else {
      hw_id                 = VPE_HW_XAV1ENC;
      m_params.codec_format = VPE_ENC_AV1;
      m_params.av1_type     = VPI_ENC_TYPE_1;
    }
    m_params.profile = const_cast<char*>(GetEnumName(AV1ProfileList, sizeof(AV1ProfileList) / sizeof(AV1ProfileList[0]), props->profile));
    if (m_params.profile == nullptr) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown AV1 profile!");
    }
    switch (props->level) {
    case 0:
      m_params.level = const_cast<char*>("auto");
      break;
    case 20:
    case 21:
    case 30:
    case 31:
    case 40:
    case 41:
    case 50:
    case 51:
    case 52:
    case 53:
    case 60:
    case 61:
    case 62:
    case 63:
      m_params.level = const_cast<char*>(std::to_string(props->level).insert(1, ".").c_str());
      break;
    default:
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unknown AV1 level!");
    }
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid encoder type!");
  }

  VpiHwCfg* hw_cfg = vpi_device_get_hw_config(m_sf->GetDevHandle(), hw_id);
  if ((props->width < hw_cfg->min_width) || (hw_cfg->max_width < props->width) || (props->height < hw_cfg->min_height) || (hw_cfg->max_height < props->height)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Unsupported frame size!");
  }

  if (props->format == XMA_VPE_FMT_TYPE) {
    m_params.input_format = props->sw_format;
  } else {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Input format not supported!");
  }

  VpiFmt* pfmt = hw_cfg->sw_formats;
  while (*pfmt != VPI_FORMAT_NONE) {
    if (*pfmt == static_cast<VpiFmt>(props->sw_format)) {
      break;
    }

    ++pfmt;
  }
  if (*pfmt == VPI_FORMAT_NONE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Input format not supported!");
  }
  vpi_device_free_hw_config(hw_cfg);

  m_params.orig_width  = props->width;
  m_params.orig_height = props->height;

  if ((props->sw_format == XMA_P010LE_FMT_TYPE) || (props->sw_format == XMA_YUV420P10LE_FMT_TYPE) || (props->sw_format == XMA_PACKED10_FMT_TYPE)) {
    m_params.bitdepth = 10;
  } else {
    m_params.bitdepth = 8;
  }

  m_params.flag         = props->flags;
  m_params.ch_index     = 0;
  m_params.enc_index    = 0;
  m_params.xav1_options = NULL;

  if (m_sf->GetAPIVersion() >= XMA_API_VERSION_1_1) {
    param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_THREADS);
    if (param) {
      m_use_thread = ((*static_cast<int32_t*>(param->value)) ? 1 : 0);
      if (m_use_thread) {
        param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_WAIT);
        if (param) {
          m_wait = *static_cast<int32_t*>(param->value);
        }
      }
    }
  }

  param = GetParameter(props->params, props->param_cnt, "extradata");
  if (param) {
    m_extra_data = *reinterpret_cast<uint8_t**>(param->value);

    param = GetParameter(props->params, props->param_cnt, "extradata_size");
    if (param) {
      m_extra_data_size = reinterpret_cast<int32_t*>(param->value);
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Extra data size parameter missing!");
    }
  }

  memset(m_param_str, 0, sizeof(m_param_str));

  char prop[64];

  m_params.slice_id = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_SLICE, XMA_ENC_SLICE_MIN_VALUE, XMA_ENC_SLICE_1, XMA_ENC_SLICE_DEFAULT);

  m_params.num_cores = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_CORES, XMA_ENC_CORES_1, XMA_ENC_CORES_2, XMA_ENC_CORES_DEFAULT, XMA_API_VERSION_1_1, XMA_API_VERSION_LATEST);
  if (m_params.num_cores == XMA_ENC_CORES_2) {
    if (props->hwencoder_type != XMA_AV1_ENCODER_TYPE || (props->hwencoder_type == XMA_AV1_ENCODER_TYPE && m_params.av1_type != VPI_ENC_TYPE_1)) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() 2-slice encoding only supported for Xilinx AV1 encoder !");
    }
    if (m_params.la_depth != 0) {
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() For 2-slice encoding, lookahead depth should be set to 0 !");
    }
    if (m_params.slice_id != XMA_ENC_SLICE_DEFAULT) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
          "InternalEncoderSession::InitParams() For 2-slice encoding, slice id needs to be set internally. Please set it to -1 !");
    }
  }
  m_params.spatial_aq = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_SPATIAL_AQ, XMA_ENC_SPATIAL_AQ_AUTO, XMA_ENC_SPATIAL_AQ_ENABLE, XMA_ENC_SPATIAL_AQ_DEFAULT);

  m_params.temporal_aq = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_TEMPORAL_AQ, XMA_ENC_TEMPORAL_AQ_AUTO, XMA_ENC_TEMPORAL_AQ_ENABLE, XMA_ENC_TEMPORAL_AQ_DEFAULT);

  m_params.enc_force_idr = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_FORCED_IDR, XMA_ENC_FORCED_IDR_DISABLE, XMA_ENC_FORCED_IDR_ENABLE, XMA_ENC_FORCED_IDR_DEFAULT);

  m_params.crf = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_CRF, XMA_ENC_PARAM_CRF_DISABLE, XMA_ENC_PARAM_CRF_ENABLE, XMA_ENC_PARAM_CRF_DEFAULT);

  m_params.dynamic_gop = SetParam(
      props->params, props->param_cnt, XMA_ENC_PARAM_DYNAMIC_GOP, XMA_ENC_DYNAMIC_GOP_AUTO, XMA_ENC_DYNAMIC_GOP_ENABLE, XMA_ENC_DYNAMIC_GOP_DEFAULT);

  m_params.tune_metrics = static_cast<TuneMetrics>(
      SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_TUNE_METRICS, XMA_ENC_TUNE_METRICS_VQ, XMA_ENC_TUNE_METRICS_VMAF, XMA_ENC_TUNE_METRICS_DEFAULT));

  m_params.qp_mode = static_cast<QpMode>(
      SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_QP_MODE, XMA_ENC_QP_MODE_AUTO, XMA_ENC_QP_MODE_UNIFORM, XMA_ENC_QP_MODE_DEFAULT));

  m_params.latency_ms = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_LATENCY_MS, XMA_ENC_LATENCY_MS_MIN, XMA_ENC_LATENCY_MS_MAX, XMA_ENC_LATENCY_MS_DEFAULT,
      XMA_API_VERSION_1_1, XMA_API_VERSION_LATEST);

  m_params.no_low_latency_b_frames = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_NO_LOWLAT_BFRAMES, XMA_ENC_NO_LOWLAT_BFRAMES_MIN,
      XMA_ENC_NO_LOWLAT_BFRAMES_ENABLE, XMA_ENC_NO_LOWLAT_BFRAMES_DEFAULT, XMA_API_VERSION_1_1_2, XMA_API_VERSION_LATEST);

  if (m_params.no_low_latency_b_frames == 0) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid value for no_low_latency_b_frames=%d", m_params.no_low_latency_b_frames);
  }

  m_params.bufsize = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_BUFSIZE, XMA_ENC_BUFSIZE_MIN, XMA_ENC_BUFSIZE_MAX, XMA_ENC_BUFSIZE_DEFAULT,
      XMA_API_VERSION_1_1, XMA_API_VERSION_LATEST);

  int8_t max_num_bframes;
  if (m_params.av1_type == VPI_ENC_TYPE_1) {
    max_num_bframes = XMA_AV1_AMD_ENC_MAX_BF_VALUE;
  } else {
    max_num_bframes = XMA_ENC_MAX_BF_VALUE;
  }
  m_params.num_b_frames = SetParam(props->params, props->param_cnt, XMA_ENC_PARAM_BF, XMA_ENC_MIN_BF_VALUE, max_num_bframes, XMA_ENC_BF_DEFAULT);

  if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_MAX_BITRATE)) != nullptr) {
    int64_t param_value = *reinterpret_cast<int64_t*>(param->value);
    if (param_value < XMA_ENC_MAX_BITRATE_MIN_VALUE || param_value > XMA_ENC_MAX_BITRATE_MAX_VALUE) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::InitParams() Invalid max_bitrate value");
    } else {
      m_params.max_bitrate = *reinterpret_cast<int64_t*>(param->value);
    }
  } else {
    m_params.max_bitrate = XMA_ENC_PARAM_MAX_BITRATE_DEFAULT;
  }

  if ((param = GetParameter(props->params, props->param_cnt, XMA_ENC_PARAM_EXPERT_OPTIONS)) != nullptr) {
    strcpy(prop, (char*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "useMemoryCompression")) != nullptr) {
    sprintf(prop, ",useMemoryCompression=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "openGop")) != nullptr) {
    sprintf(prop, ",openGop=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "intraRefreshPeriod")) != nullptr) {
    sprintf(prop, ",intraRefreshPeriod=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "maxM")) != nullptr) {
    sprintf(prop, ",maxM=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "bll")) != nullptr) {
    sprintf(prop, ",bll=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "strictIdr")) != nullptr) {
    sprintf(prop, ",strictIdr=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vbvSize")) != nullptr) {
    sprintf(prop, ",vbvSize=%u", *(uint32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vbvTarget")) != nullptr) {
    sprintf(prop, ",vbvTarget=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "needRecon")) != nullptr) {
    sprintf(prop, ",needRecon=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "numVTiles")) != nullptr) {
    sprintf(prop, ",numVTiles=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "numHTiles")) != nullptr) {
    sprintf(prop, ",numHTiles=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "useBothEncoderInstances")) != nullptr) {
    sprintf(prop, ",useBothEncoderInstances=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "tune")) != nullptr) {
    sprintf(prop, ",tune=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "hdr")) != nullptr) {
    sprintf(prop, ",hdr=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "aqEnable")) != nullptr) {
    sprintf(prop, ",aqEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "qpMapEnable")) != nullptr) {
    sprintf(prop, ",qpMapEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "dynamicGop")) != nullptr) {
    sprintf(prop, ",dynamicGop=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "mbtreeEnable")) != nullptr) {
    sprintf(prop, ",mbtreeEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "dynamicRoundingEnable")) != nullptr) {
    sprintf(prop, ",dynamicRoundingEnable=%s", (*(int32_t*) param->value) ? "true" : "false");
    strcat(m_param_str, prop);
  }

  if ((param = GetParameter(props->params, props->param_cnt, "vceVertSearchRange")) != nullptr) {
    sprintf(prop, ",vceVertSearchRange=%d", *(int32_t*) param->value);
    strcat(m_param_str, prop);
  }

  if (m_param_str != NULL) {
    m_params.xav1_options = m_param_str;
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::UpdateParams(FramePtr frame) {
  if (frame) {
    SideDataPtr sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS);
    if (sd) {
      XmaBufferType buffer_type = sd->GetBufferType();
      if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
            "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS!");
      }
      if (m_sf->GetAPIVersion() < XMA_API_VERSION_1_1_2) {
        if (sd->GetSize() != sizeof(XmaDynamicEncParams)) {
          THROW_RUNTIME_EXCEPTION(
              m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::UpdateParams() Invalid size for XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS side data!");
        }
        XmaDynamicEncParams* dyn_params = reinterpret_cast<XmaDynamicEncParams*>(sd->GetHostPtr());
        if (dyn_params->is_spatial_aq_gain_changed) {
          m_params.spatial_aq_gain = dyn_params->spatial_aq_gain;
        }
        if (dyn_params->is_temporal_aq_gain_changed) {
          m_params.temporal_aq_gain = dyn_params->temporal_aq_gain;
        }
        if (dyn_params->is_temporal_mode_changed) {
          m_params.temporal_aq = dyn_params->temporal_aq_mode;
        }
        if (dyn_params->is_spatial_mode_changed) {
          m_params.spatial_aq = dyn_params->spatial_aq_mode;
        }
        if (dyn_params->is_bit_rate_changed) {
          m_params.bit_rate = dyn_params->bit_rate_kbps;
        }
        if (dyn_params->is_b_frames_changed) {
          m_params.num_b_frames = dyn_params->num_b_frames;
        }
        if (dyn_params->is_min_qp_changed) {
          m_params.min_qp = dyn_params->min_qp;
        }
        if (dyn_params->is_max_qp_changed) {
          m_params.max_qp = dyn_params->max_qp;
        }
      } else {
        if (sd->GetSize() != sizeof(XmaDynamicEncParams_v2)) {
          THROW_RUNTIME_EXCEPTION(
              m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::UpdateParams() Invalid size for XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS side data!");
        }
        XmaDynamicEncParams_v2* dyn_params = reinterpret_cast<XmaDynamicEncParams_v2*>(sd->GetHostPtr());
        if (dyn_params->v1.is_spatial_aq_gain_changed) {
          m_params.spatial_aq_gain = dyn_params->v1.spatial_aq_gain;
        }
        if (dyn_params->v1.is_temporal_aq_gain_changed) {
          m_params.temporal_aq_gain = dyn_params->v1.temporal_aq_gain;
        }
        if (dyn_params->v1.is_temporal_mode_changed) {
          m_params.temporal_aq = dyn_params->v1.temporal_aq_mode;
        }
        if (dyn_params->v1.is_spatial_mode_changed) {
          m_params.spatial_aq = dyn_params->v1.spatial_aq_mode;
        }
        if (dyn_params->v1.is_bit_rate_changed) {
          m_params.bit_rate = dyn_params->v1.bit_rate_kbps;
        }
        if (dyn_params->is_min_bit_rate_changed) {
          m_params.min_bitrate = dyn_params->min_bit_rate_kbps;
        }
        if (dyn_params->is_max_bit_rate_changed) {
          m_params.max_bitrate = dyn_params->max_bit_rate_kbps;
        }
        if (dyn_params->v1.is_b_frames_changed) {
          m_params.num_b_frames = dyn_params->v1.num_b_frames;
        }
        if (dyn_params->is_qp_changed) {
          m_params.qp = dyn_params->qp;
        }
        if (dyn_params->v1.is_min_qp_changed) {
          m_params.min_qp = dyn_params->v1.min_qp;
        }
        if (dyn_params->v1.is_max_qp_changed) {
          m_params.max_qp = dyn_params->v1.max_qp;
        }
        if (dyn_params->is_qp_i_offset_changed) {
          m_params.qp_i_offset = dyn_params->qp_i_offset;
        }
        if (dyn_params->is_qp_b_offset_changed) {
          m_params.qp_b_offset = dyn_params->qp_b_offset;
        }
      }
    }

    sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_VUI_PARAMS);
    if (sd) {
      XmaBufferType buffer_type = sd->GetBufferType();
      if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
            "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_VUI_PARAMS!");
      }
      memcpy(&m_params.vui_param, sd->GetHostPtr(), sizeof(XmaVuiParams));
    }

    XmaAPIVersion api_version = m_sf->GetAPIVersion();
    if (api_version < XMA_API_VERSION_1_1) {
      sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS);
      if (sd) {
        XmaBufferType buffer_type = sd->GetBufferType();
        if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
              "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_HDR10_PARAMS!");
        }
        XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());
        if (hdr_params && (sd->GetSize() == sizeof(XmaHdrParams))) {
          memcpy(&m_params.mdcv, &hdr_params->mdcv, sizeof(XmaMasterDisColVol));
          memcpy(&m_params.cll, &hdr_params->cll, sizeof(XmaContentLightLevel));
          memcpy(&m_params.atc, &hdr_params->atc, sizeof(XmaATCInfo));
        }
      }
    } else {
      sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_SEI_MASTERING_DISPLAY_COLOR_VOLUME);
      uint8_t* host;
      if (sd) {
        XmaBufferType buffer_type = sd->GetBufferType();
        if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
              "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_SEI_MASTERING_DISPLAY_COLOR_VOLUME!");
        }
        host = sd->GetHostPtr();
        if (host && (sd->GetSize() == sizeof(XmaMasterDisColVol))) {
          memcpy(&m_params.mdcv, host, sizeof(XmaMasterDisColVol));
        }
      }
      sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_SEI_CONTENT_LIGHT_LEVEL);
      if (sd) {
        XmaBufferType buffer_type = sd->GetBufferType();
        if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
              "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_SEI_CONTENT_LIGHT_LEVEL!");
        }
        host = sd->GetHostPtr();
        if (host && (sd->GetSize() == sizeof(XmaContentLightLevel))) {
          memcpy(&m_params.cll, host, sizeof(XmaContentLightLevel));
        }
      }
      sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_SEI_ATC);
      if (sd) {
        XmaBufferType buffer_type = sd->GetBufferType();
        if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION,
              "InternalEncoderSession::UpdateParams() Invalid side data buffer type for side data XMA_FRAME_SIDE_DATA_SEI_ATC!");
        }
        host = sd->GetHostPtr();
        if (host && (sd->GetSize() == sizeof(XmaATCInfo))) {
          memcpy(&m_params.atc, host, sizeof(XmaATCInfo));
        }
      }
    }
  }
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::StartEncoder(FramePtr& frame) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder()");

  if (m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Encoder already initialized!");
  }

  if (frame) {
    SideDataPtr sd = frame->GetSideData(XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS);
    if (sd) {
      frame = std::make_shared<InternalFrame>(m_sf, m_log, frame);
      frame->RemoveSideDataType(XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS);
    }
  }

  m_handle = enc_handle_t(vpi_encoder_init(&m_params), [](VpiEncHandle handle) { vpi_encoder_close(handle); });
  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Failed to initialize encoder!");
  }

  if (vpi_encoder_open(m_handle.get()) < 0) {
    m_handle.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() Unable to open encoder for processing!");
  }

  VpiEncContext* enc_ctx      = static_cast<VpiEncContext*>(m_handle.get());
  Xav1Context*   enc_xav1_ctx = static_cast<Xav1Context*>(enc_ctx->vpi_encoder_handle);
  m_max_frames_processing     = enc_xav1_ctx->max_frames_delay;
  if (m_params.la_depth == 0) {
    m_max_frames_processing = 1;
  }

  if (m_extra_data && m_extra_data_size) {
    uint8_t* extradata_buffer = reinterpret_cast<uint8_t*>(vpi_encoder_get_extradata(m_handle.get(), m_extra_data_size));
    memcpy(m_extra_data, extradata_buffer, *m_extra_data_size);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::StartEncoder() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Push_v1_0(FramePtr frame) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0()");

  lock_t guard(m_mutex);

  if (!m_handle.get()) {
    StartEncoder(frame);
  }

  if (!frame) {
    if (!m_eos_pushed) {
      m_eos_pushed = true;
      int ret;
      if ((ret = vpi_encoder_put_frame(m_handle.get(), nullptr)) != 0) {
        THROW_RUNTIME_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() Failed to send end of file to encoder (%s)!", vpi_error_str(ret));
      }
    }
    ++m_frames_processing;

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() - Returning success");
    return XMA_SUCCESS;
  }

  FramePropertiesPtr props = frame->GetProps();
  if ((m_params.width != props->GetWidth()) || (m_params.height != props->GetHeight())) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() Unsupported properties change!");
  }
  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() Out of memory!");
  }
  InitVpiFrm(pic);

  if (AddFrameToVpiFrm(m_sf, m_log, frame, pic) != XMA_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() Failed to add XmaFrame to VpiFrm!");
  }

  pic->poc = 0;
  pic->pts = frame->GetPts();
  pic->dts = 0;

  if (vpi_encoder_put_frame(m_handle.get(), pic) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() Failed to send frame to encoder!");
  }
  ++m_frames_processing;

  int32_t ret = PullPacket();
  if (ret == XMA_EOS) {
    m_queue.push_back(nullptr);
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() - Returning success");
    return XMA_SUCCESS;
  } else if (ret == XMA_RESEND_AND_RECV) {
    if (m_frames_processing >= m_max_frames_processing) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() - Returning success");
      return XMA_SUCCESS;
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() - Returning send more data");
      return XMA_SEND_MORE_DATA;
    }
  } else {
    m_queue.push_back(std::move(m_buf));
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_0() - Returning success");
    return XMA_SUCCESS;
  }
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Pull_v1_0(IntDataBufferPtr& buf) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0()");

  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() Encoder not initialized!");
  }

  lock_t guard(m_mutex);

  if (!m_queue.empty()) {
    buf = m_queue.front();
    m_queue.pop_front();
    if (buf == nullptr) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() - Returning EOS");
      return XMA_EOS;
    }
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() - Returning success");
    return XMA_SUCCESS;
  }

  int ret = PullPacket();
  if (ret == XMA_EOS) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() - Returning EOS");
    return ret;
  } else if (ret == XMA_RESEND_AND_RECV) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() - Returning resend and recv");
    return XMA_RESEND_AND_RECV;
  }
  buf = std::move(m_buf);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_0() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Push_v1_1(FramePtr frame) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1()");

  lock_t guard(m_mutex);

  if (m_frame) {
    if (!m_wait) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1() - Returning try again");
      return XMA_TRY_AGAIN;
    } else {
      do {
        m_wait_enqueue_cond.wait(guard);
        if (m_exception.HasException()) {
          m_exception.RethrowException();
        }
        if (m_stop) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
          return XMA_ERROR_TIMEOUT;
        }
      } while (m_frame);
    }
  }

  if (frame) {
    FramePropertiesPtr props = frame->GetProps();
    if ((m_params.width != props->GetWidth()) || (m_params.height != props->GetHeight()) || m_resetting) {
      m_resetting     = true;
      m_params.width  = props->GetWidth();
      m_params.height = props->GetHeight();
      m_thread_enqueue_cond.notify_all();
      if (!m_use_thread) {
        m_eos_pushed = true;
        FramePtr eos_frame;
        PushFrame_v1_1(eos_frame);
      }
      if (m_wait) {
        do {
          m_wait_enqueue_cond.wait(guard);
          if (m_exception.HasException()) {
            m_exception.RethrowException();
          }
          if (m_stop) {
            m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
            return XMA_ERROR_TIMEOUT;
          }
        } while (m_resetting);
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1() - Returning try again");
        return XMA_TRY_AGAIN;
      }
    }
    UpdateParams(frame);

    if (!m_handle.get()) {
      StartEncoder(frame);
    }
  } else {
    m_eos_pushed = true;
  }

  if (m_use_thread) {
    if (frame) {
      frame->IncRef();
    }
    m_frame = std::move(frame);
  } else {
    PushFrame_v1_1(frame);
  }
  m_thread_enqueue_cond.notify_one();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Push_v1_1() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Pull_v1_1(IntDataBufferPtr& buf) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1()");

  lock_t guard(m_mutex);

  if (!m_use_thread) {
    int32_t ret;
    while (true) {
      ret = Process_ULL(buf);
      if (ret == XMA_SUCCESS) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning success");
        return XMA_SUCCESS;
      } else if (ret == XMA_EOS) {
        if (m_resetting) {
          m_eos_pushed   = false;
          m_end_recieved = false;
          m_resetting    = false;
          m_handle.reset();
          m_wait_enqueue_cond.notify_one();
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning resend and receive");
          return XMA_RESEND_AND_RECV;
        } else {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning EOS");
          return XMA_EOS;
        }
      } else if (!m_wait) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning resend and receive");
        return XMA_RESEND_AND_RECV;
      }
      m_thread_enqueue_cond.wait(guard);
      if (m_stop) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
        return XMA_ERROR_TIMEOUT;
      }
    }
  }

  if ((!m_buf) && ((!m_end_recieved) || (m_resetting))) {
    if (m_wait) {
      do {
        m_wait_dequeue_cond.wait(guard);
        if (m_exception.HasException()) {
          m_exception.RethrowException();
        }
        if (m_stop) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning XMA_ERROR_TIMEOUT: time alloted for call exceeded");
          return XMA_ERROR_TIMEOUT;
        }
      } while ((!m_buf) && (!m_end_recieved));
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning resend and receive");
      return XMA_RESEND_AND_RECV;
    }
  }

  buf = std::move(m_buf);
  if (m_use_thread) {
    m_thread_dequeue_cond.notify_one();
  }

  if (buf) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning success");
    return XMA_SUCCESS;
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Pull_v1_1() - Returning EOS");
    return XMA_EOS;
  }
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::PushFrame_v1_1(FramePtr& frame) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PushFrame_v1_1()");

  int     ret;
  VpiFrm* pic = nullptr;

  if (frame) {
    if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PushFrame_v1_1() Out of memory!");
    }
    InitVpiFrm(pic);

    if (AddFrameToVpiFrm(m_sf, m_log, frame, pic) != XMA_SUCCESS) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PushFrame_v1_1() Failed to add XmaFrame to VpiFrm!");
    }

    pic->poc = 0;
    pic->pts = frame->GetPts();
    pic->dts = 0;
  }

  if ((ret = vpi_encoder_put_frame(m_handle.get(), pic)) < 0) {
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PushFrame_v1_1() Failed to send frame to encoder (%s)!", vpi_error_str(ret));
  }
  if (frame) {
    if (m_use_thread) {
      frame->DecRef();
    }
  }
  frame.reset();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PushFrame_v1_1() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
IntDataBufferPtr InternalEncoderSession::Process_v1_1() {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1()");

  if (m_end_recieved) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() - Returning EOS");
    return nullptr;
  }

  int size = 0;
  vpi_encoder_get_packet_size(m_handle.get(), &size);

  IntDataBufferPtr buf = std::make_shared<InternalDataBuffer>(m_sf, m_log, size, false);

  VpiPkt pkt;
  memset(&pkt, 0, sizeof(VpiPkt));

  while (true) {
    pkt.buffer = static_cast<uint8_t*>(buf->GetBuffer());
    pkt.size   = buf->GetSize();

    int ret;
    if ((ret = vpi_encoder_get_packet(m_handle.get(), &pkt)) != 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() Failed to get packet (%s)!", vpi_error_str(ret));
    }

    if (m_eos_pushed && (pkt.size == 0) && pkt.is_eof) {
      m_end_recieved = true;
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() - Returning EOS");
      return nullptr;
    }
    if (pkt.size == 0) {
      if (m_eos_pushed) {
        if ((ret = vpi_encoder_put_frame(m_handle.get(), nullptr)) < 0) {
          THROW_RUNTIME_EXCEPTION(
              m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() Failed to send frame to encoder (%s)!", vpi_error_str(ret));
        }
        usleep(100);
        continue;
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() - Returning send more data");
        return nullptr;
      }
    }
    buf->SetSize(pkt.size);
    buf->SetPts(pkt.pts);

    if (m_params.codec_format != VPE_ENC_AV1_T2) {
      m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder PSNR Y=%f U=%f V=%f", pkt.psnr[0], pkt.psnr[1], pkt.psnr[2]);
      m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder SSIM Y=%f U=%f V=%f", pkt.ssim[0], pkt.ssim[1], pkt.ssim[2]);
    }

    if (m_eos_pushed) {
      if (pkt.is_eof) {
        m_end_recieved = true;
      }
    }
    break;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_v1_1() - Returning success");
  return buf;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::Process_ULL(IntDataBufferPtr& buf) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_ULL()");

  if (m_handle) {
    buf = Process_v1_1();
  }
  if (buf) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_ULL() - Returning success");
    return XMA_SUCCESS;
  }
  if (m_end_recieved) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_ULL() - Returning EOS");
    return XMA_EOS;
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Process_ULL() - Returning resend and receive");
  return XMA_RESEND_AND_RECV;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::GetPacket(IntDataBufferPtr& buf) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket()");
  VpiPkt pkt;
  memset(&pkt, 0, sizeof(VpiPkt));
  for (int32_t retry_count = XMA_ENCODER_MAX_RETRY_COUNT; retry_count >= 0; --retry_count) {
    pkt.buffer = static_cast<uint8_t*>(buf->GetBuffer());
    pkt.size   = buf->GetSize();

    if (vpi_encoder_get_packet(m_handle.get(), &pkt) != 0) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Failed to get packet!");
    }

    if (pkt.size == 0) {
      if (m_eos_pushed) {
        if (pkt.is_eof) {
          m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning end of file");
          return XMA_EOS;
        }
        if (retry_count > 0) {
          if (vpi_encoder_put_frame(m_handle.get(), nullptr) < 0) {
            THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Failed to send NULL to encoder!");
          }
          usleep(static_cast<uint32_t>(1000000L * m_params.rate_denom / (m_params.rate_num * XMA_ENCODER_RETRY_DELAY)));
          continue;
        } else {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() Recieve failed during flush!");
        }
      } else if ((m_wait) && (m_frames_processing >= m_max_frames_processing)) {
        if (retry_count > 0) {
          usleep(static_cast<uint32_t>(1000000L * m_params.rate_denom / (m_params.rate_num * XMA_ENCODER_RETRY_DELAY)));
          continue;
        } else {
          THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetXAV1Packet() Timeout waiting for encoder to get packet!");
        }
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning resend and recv");
        return XMA_RESEND_AND_RECV;
      }
    } else {
      break;
    }
  }

  buf->SetSize(pkt.size);
  buf->SetPts(pkt.pts);

  if (m_params.codec_format != VPE_ENC_AV1_T2) {
    m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder PSNR Y=%f U=%f V=%f", pkt.psnr[0], pkt.psnr[1], pkt.psnr[2]);
    m_log->LogMsg(XMA_INFO_LOG, XMA_ENCODER_SESSION, "Encoder SSIM Y=%f U=%f V=%f", pkt.ssim[0], pkt.ssim[1], pkt.ssim[2]);
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::GetPacket() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalEncoderSession::PullPacket() {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, m_instance_name);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PullPacket()");
  int size = 0;
  vpi_encoder_get_packet_size(m_handle.get(), &size);

  if ((!m_buf) || (m_buf->GetSize() < size)) {
    m_buf = std::make_shared<InternalDataBuffer>(m_sf, m_log, size, false);
  }

  int32_t ret = GetPacket(m_buf);
  if (ret != XMA_SUCCESS) {
    return ret;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::PullPacket() - Returning success");
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
void InternalEncoderSession::Thread(InternalEncoderSession* session) {
  LogAmaInstanceMgmt ama_log_env(LOG_AMA_ACCEL_ENC, session->m_instance_name);
  session->m_log->LogMsg(XMA_DEBUG_LOG, XMA_ENCODER_SESSION, "InternalEncoderSession::Thread()");

  lock_t lock(session->m_mutex);
  while ((!session->m_stop) && ((!session->m_end_recieved) || (session->m_resetting))) {
    while ((!session->m_frame) && (!session->m_eos_pushed) && (!session->m_resetting)) {
      session->m_thread_enqueue_cond.wait(lock);
      if (session->m_stop) {
        return;
      }
    }

    FramePtr frame;
    if (!session->m_resetting) {
      frame = std::move(session->m_frame);
      session->m_wait_enqueue_cond.notify_one();
    } else {
      session->m_eos_pushed = true;
    }
    lock.unlock();

    try {
      session->PushFrame_v1_1(frame);
      IntDataBufferPtr buf = session->Process_v1_1();

      lock.lock();
      if (buf) {
        while (session->m_buf) {
          session->m_thread_dequeue_cond.wait(lock);
          if (session->m_stop) {
            return;
          }
        }
        session->m_buf = std::move(buf);
        session->m_wait_dequeue_cond.notify_one();
      } else if (session->m_end_recieved) {
        if (session->m_resetting) {
          session->m_eos_pushed   = false;
          session->m_end_recieved = false;
          session->m_resetting    = false;
          session->m_handle.reset();
          session->m_wait_enqueue_cond.notify_all();
        } else {
          while (session->m_buf) {
            session->m_thread_dequeue_cond.wait(lock);
            if (session->m_stop) {
              return;
            }
          }
          session->m_wait_dequeue_cond.notify_one();
          return;
        }
      }
    } catch (ThreadThrowable& exception) {
      session->m_exception.SetException(exception.CreateClone());
      session->m_wait_enqueue_cond.notify_all();
      session->m_wait_dequeue_cond.notify_all();
      return;
    } catch (...) {
      session->m_exception.SetException(std::make_shared<UnknownException>());
      session->m_wait_enqueue_cond.notify_all();
      session->m_wait_dequeue_cond.notify_all();
      return;
    }
  }
}

} // namespace xma
